@@ -3,7 +3,9 @@
 import os
 import math
 import warnings
+from typing import Optional, List
 
+from argus import types
 from argus.engine import State
 from argus.callbacks.callback import Callback
 from argus.metrics.metric import init_better
@@ -16,8 +18,8 @@ class Checkpoint(Callback):
     each epoch.
 
     Args:
-        dir_path (str, optional): Directory to save checkpoints. The
-            desired directory will be created if it does not exist.
+        dir_path (str or :class:`pathlib.Path`): Directory to save checkpoints.
+            The desired directory will be created if it does not exist.
             Defaults to ''.
         file_format (str, optional): Model saving filename format. Any
             valid value names from the model State may be used. Defaults to
@@ -29,22 +31,24 @@ class Checkpoint(Callback):
             checkpoint saves. Defaults to 1.
         save_after_exception (bool, optional): Save the model checkpoint
             after an exception occurs. Defaults to False.
+        optimizer_state (bool): Save optimizer state. Defaults to False.
 
     """
 
     def __init__(self,
-                 dir_path='',
-                 file_format='model-{epoch:03d}-{train_loss:.6f}.pth',
-                 max_saves=None,
-                 period=1,
-                 save_after_exception=False):
+                 dir_path: types.Path = '',
+                 file_format: str = 'model-{epoch:03d}-{train_loss:.6f}.pth',
+                 max_saves: Optional[int] = None,
+                 period: int = 1,
+                 save_after_exception: bool = False,
+                 optimizer_state: bool = False):
         if not (max_saves is None or max_saves > 0):
             raise ValueError("max_saves should be positive or 'None'")
 
         self.dir_path = dir_path
         self.file_format = file_format
         self.max_saves = max_saves
-        self.saved_files_paths = []
+        self.saved_files_paths: List[types.Path] = []
         if self.dir_path:
             if not os.path.exists(dir_path):
                 os.makedirs(dir_path)
@@ -52,18 +56,19 @@ class Checkpoint(Callback):
                 warnings.warn(f"Directory '{dir_path}' already exists")
         self.period = period
         self.save_after_exception = save_after_exception
-        self.epochs_since_last_save = 0
+        self.optimizer_state = optimizer_state
+        self.epochs_since_last_save: int = 0
 
-    def save_model(self, state: State, file_path):
+    def save_model(self, state: State, file_path: types.Path):
         """Save model to file.
 
         Override the method if you need custom checkpoint saving.
 
         Args:
             state (:class:`argus.engine.State`): State.
-            file_path (str): Checkpoint file path.
+            file_path (str or :class:`pathlib.Path`): Checkpoint file path.
         """
-        state.model.save(file_path)
+        state.model.save(file_path, optimizer_state=self.optimizer_state)
 
     def _format_file_path(self, state: State):
         format_state = {'epoch': state.epoch, **state.metrics}
@@ -110,17 +115,18 @@ class MonitorCheckpoint(Checkpoint):
     metric available in the model State.
 
     Args:
-        dir_path (str, optional): Directory to save checkpoints. The
-            desired directory will be created if it does not exist.
+        dir_path (str or :class:`pathlib.Path`): Directory to save checkpoints.
+            The desired directory will be created if it does not exist.
             Defaults to ''.
         file_format (str, optional): Model saving filename format. Any
             valid value names from the model State may be used. Defaults to
             'model-{epoch:03d}-{monitor:.6f}.pth'.
-        max_saves ([type], optional): Number of last saved models to keep.
+        max_saves (int, optional): Number of last saved models to keep.
             Should be positive. If None - save all models. Defaults to
             None.
         save_after_exception (bool, optional): Save the model checkpoint
             after an exception occurs. Defaults to False.
+        optimizer_state (bool): Save optimizer state. Defaults to False.
         monitor (str, optional): Metric name to monitor. It should be
             prepended with *val_* for the metric value on validation data
             and *train_* for the metric value on the date from the train
@@ -135,12 +141,13 @@ class MonitorCheckpoint(Checkpoint):
     """
 
     def __init__(self,
-                 dir_path='',
-                 file_format='model-{epoch:03d}-{monitor:.6f}.pth',
-                 max_saves=None,
-                 save_after_exception=False,
-                 monitor='val_loss',
-                 better='auto'):
+                 dir_path: types.Path = '',
+                 file_format: str = 'model-{epoch:03d}-{monitor:.6f}.pth',
+                 max_saves: Optional[int] = None,
+                 save_after_exception: bool = False,
+                 optimizer_state: bool = False,
+                 monitor: str = 'val_loss',
+                 better: str = 'auto'):
         if not monitor.startswith('val_') and not monitor.startswith('train_'):
             raise ValueError("monitor should be prepended with 'val_' or 'train_'")
 
@@ -148,7 +155,8 @@ class MonitorCheckpoint(Checkpoint):
                          file_format=file_format,
                          max_saves=max_saves,
                          period=1,
-                         save_after_exception=save_after_exception)
+                         save_after_exception=save_after_exception,
+                         optimizer_state=optimizer_state)
         self.monitor = monitor
         self.better, self.better_comp, self.best_value = init_better(
             better, monitor)
