@@ -4,45 +4,36 @@ import os
 import csv
 import logging
 from datetime import datetime
+from typing import Optional, Union, List, IO
 
+from argus import types
 from argus.engine import State
 from argus.callbacks.callback import Callback, on_epoch_complete
 
 
-def _format_lr_to_str(lr, precision=5):
+def _format_lr_to_str(lr: Union[float, List[float]],
+                      precision: int = 5) -> str:
     if isinstance(lr, (list, tuple)):
-        lr = [f'{l:.{precision}g}' for l in lr]
-        lr = "[" + ", ".join(lr) + "]"
+        str_lrs = [f'{l:.{precision}g}' for l in lr]
+        str_lr = "[" + ", ".join(str_lrs) + "]"
     else:
-        lr = f'{lr:.{precision}g}'
-    return lr
+        str_lr = f'{lr:.{precision}g}'
+    return str_lr
 
 
 @on_epoch_complete
-def metrics_logging(state: State, train=False, print_epoch=True):
-    if train:
-        epoch_name = 'Train'
-        prefix = 'train_'
-    else:
-        epoch_name = 'Validation'
-        prefix = 'val_'
+def default_logging(state: State):
+    message = f"{state.phase} - epoch: {state.epoch}"
 
-    if print_epoch:
-        train_epoch = state.epoch
-        message = [f"{epoch_name} - Epoch: {train_epoch}"]
-    else:
-        message = [epoch_name]
-
-    if train:
+    if state.phase == 'train':
         lr = state.model.get_lr()
-        lr = _format_lr_to_str(lr)
-        message.append(f'LR: {lr}')
+        message += f', lr: {_format_lr_to_str(lr)}'
 
+    prefix = f"{state.phase}_" if state.phase else ''
     for metric_name, metric_value in state.metrics.items():
-        if not metric_name.startswith(prefix):
-            continue
-        message.append(f"{metric_name}: {metric_value:.7g}")
-    state.logger.info(", ".join(message))
+        if metric_name.startswith(prefix):
+            message += f", {metric_name}: {metric_value:.7g}"
+    state.logger.info(message)
 
 
 class LoggingToFile(Callback):
@@ -64,15 +55,16 @@ class LoggingToFile(Callback):
 
     """
 
-    def __init__(self, file_path,
-                 create_dir=True,
-                 formatter='[%(asctime)s][%(levelname)s]: %(message)s',
-                 append=False):
+    def __init__(self,
+                 file_path: types.Path,
+                 create_dir: bool = True,
+                 formatter: str = '[%(asctime)s][%(levelname)s]: %(message)s',
+                 append: bool = False):
         self.file_path = file_path
         self.create_dir = create_dir
         self.formatter = logging.Formatter(formatter)
         self.append = append
-        self.file_handler = None
+        self.file_handler: Optional[logging.FileHandler] = None
 
     def start(self, state: State):
         if self.create_dir:
@@ -87,7 +79,8 @@ class LoggingToFile(Callback):
         state.logger.addHandler(self.file_handler)
 
     def complete(self, state: State):
-        state.logger.removeHandler(self.file_handler)
+        if self.file_handler is not None:
+            state.logger.removeHandler(self.file_handler)
 
     def catch_exception(self, state: State):
         self.complete(state)
@@ -112,24 +105,21 @@ class LoggingToCSV(Callback):
 
     """
 
-    def __init__(self, file_path,
-                 create_dir=True,
-                 separator=',',
-                 write_header=True,
-                 append=False):
+    def __init__(self,
+                 file_path: types.Path,
+                 create_dir: bool = True,
+                 separator: str = ',',
+                 write_header: bool = True,
+                 append: bool = False):
         self.file_path = file_path
         self.separator = separator
         self.write_header = write_header
         self.append = append
-        self.csv_file = None
+        self.csv_file: Optional[IO] = None
         self.create_dir = create_dir
 
     def start(self, state: State):
-        if self.append:
-            file_mode = 'a'
-        else:
-            file_mode = 'w'
-
+        file_mode = 'a' if self.append else 'w'
         if self.create_dir:
             dir_path = os.path.dirname(self.file_path)
             if dir_path:
@@ -138,6 +128,8 @@ class LoggingToCSV(Callback):
         self.csv_file = open(self.file_path, file_mode, newline='')
 
     def epoch_complete(self, state: State):
+        if self.csv_file is None:
+            return
         lr = state.model.get_lr()
         fields = {
             'time': str(datetime.now()),
