@@ -2,7 +2,10 @@
 
 It enables the PyTorch lr_schedulers to be used as normal argus Callbacks.
 """
+from typing import Optional, Callable, Iterable, Any, Union, List
+
 import torch
+from torch.optim import Optimizer
 from torch.optim import lr_scheduler as _scheduler
 
 from argus.engine import State
@@ -11,22 +14,26 @@ from argus.metrics.metric import init_better
 
 
 class LRScheduler(Callback):
-    def __init__(self, scheduler_factory, step_on_iteration=False):
+    def __init__(self,
+                 scheduler_factory: Callable[[Optimizer], Any],
+                 step_on_iteration: bool = False):
         self.scheduler_factory = scheduler_factory
         self.step_on_iteration = step_on_iteration
-        self.scheduler = None
+        self.scheduler: Optional[Any] = None  # Any that has .step() method
 
     def start(self, state: State):
         if self.scheduler is None:
             self.scheduler = self.scheduler_factory(state.model.optimizer)
 
     def epoch_complete(self, state: State):
-        if not self.step_on_iteration:
-            self.scheduler.step()
+        if self.scheduler is not None:
+            if not self.step_on_iteration:
+                self.scheduler.step()
 
     def iteration_complete(self, state: State):
-        if self.step_on_iteration:
-            self.scheduler.step()
+        if self.scheduler is not None:
+            if self.step_on_iteration:
+                self.scheduler.step()
 
 
 class LambdaLR(LRScheduler):
@@ -38,15 +45,21 @@ class LambdaLR(LRScheduler):
     Args:
         lr_lambda (function or list of functions): Lambda function for the
             learning rate factor computation.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
-    def __init__(self, lr_lambda, step_on_iteration=False):
+    def __init__(self,
+                 lr_lambda: Union[Callable[[int], float],
+                                  List[Callable[[int], float]]],
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.LambdaLR(opt,
-                                            lr_lambda),
+                                            lr_lambda,
+                                            last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -59,16 +72,22 @@ class StepLR(LRScheduler):
     Args:
         step_size (int): Period of learning rate update in epochs.
         gamma (float, optional): Multiplicative factor. Defaults to 0.1.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
-    def __init__(self, step_size, gamma=0.1, step_on_iteration=False):
+    def __init__(self,
+                 step_size: int,
+                 gamma: float = 0.1,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.StepLR(opt,
                                           step_size,
-                                          gamma=gamma),
+                                          gamma=gamma,
+                                          last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -79,18 +98,24 @@ class MultiStepLR(LRScheduler):
     Multiply learning rate by a given factor on each epoch from a given list.
 
     Args:
-        milestones (list of ints): List of epochs number to perform lr step.
+        milestones (list of int): List of epochs number to perform lr step.
         gamma (float, optional): Multiplicative factor. Defaults to 0.1.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
-    def __init__(self, milestones, gamma=0.1, step_on_iteration=False):
+    def __init__(self,
+                 milestones: Iterable[int],
+                 gamma: float = 0.1,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.MultiStepLR(opt,
                                                milestones,
-                                               gamma=gamma),
+                                               gamma=gamma,
+                                               last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -102,15 +127,20 @@ class ExponentialLR(LRScheduler):
 
     Args:
         gamma (float, optional): Multiplicative factor. Defaults to 0.1.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
-    def __init__(self, gamma, step_on_iteration=False):
+    def __init__(self,
+                 gamma: float,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.ExponentialLR(opt,
-                                                 gamma),
+                                                 gamma,
+                                                 last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -124,16 +154,22 @@ class CosineAnnealingLR(LRScheduler):
     Args:
         T_max (int): Max number of epochs or iterations.
         eta_min (float, optional): Min learning rate. Defaults to 0.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
-    def __init__(self, T_max, eta_min=0, step_on_iteration=False):
+    def __init__(self,
+                 T_max: int,
+                 eta_min: float = 0,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.CosineAnnealingLR(opt,
                                                      T_max,
-                                                     eta_min=eta_min),
+                                                     eta_min=eta_min,
+                                                     last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -165,7 +201,7 @@ class ReduceLROnPlateau(LRScheduler):
             Defaults to 'rel'.
         cooldown (int, optional): Number of epochs to wait before resuming
             normal operation after lr has been updated. Defaults to 0.
-        min_lr (float or list of floats, optional): Min learning rate.
+        min_lr (float or list of float, optional): Min learning rate.
             Defaults to 0.
         eps (float, optional): Min significant learning rate update.
             Defaults to 1e-8.
@@ -194,7 +230,8 @@ class ReduceLROnPlateau(LRScheduler):
         )
 
     def epoch_complete(self, state: State):
-        self.scheduler.step(metrics=state.metrics[self.monitor])
+        if self.scheduler is not None:
+            self.scheduler.step(metrics=state.metrics[self.monitor])
 
 
 class CyclicLR(LRScheduler):
@@ -204,8 +241,8 @@ class CyclicLR(LRScheduler):
     learning rate policy.
 
     Args:
-        base_lr (float or list of floats): Initial learning rate.
-        max_lr (float or list of floats): Max learning rate.
+        base_lr (float or list of float): Initial learning rate.
+        max_lr (float or list of float): Max learning rate.
         step_size_up (int, optional): Increase phase duration in epochs or iterations.
             Defaults to 2000.
         step_size_down (int, optional): Decrease phase duration in epochs or iterations.
@@ -221,30 +258,32 @@ class CyclicLR(LRScheduler):
         cycle_momentum (bool, optional): Momentum is cycled inversely
             to learning rate between 'base_momentum' and 'max_momentum'.
             Defaults to True.
-        base_momentum (float or list of floats, optional): Lower momentum
+        base_momentum (float or list of float, optional): Lower momentum
             boundaries in the cycle for each parameter group.
             Defaults to 0.8.
-        max_momentum (float or list of floats, optional): Upper momentum
+        max_momentum (float or list of float, optional): Upper momentum
             boundaries in the cycle for each parameter group.
             Defaults to 0.9.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to True.
 
     """
 
     def __init__(self,
-                 base_lr,
-                 max_lr,
-                 step_size_up=2000,
-                 step_size_down=None,
-                 mode='triangular',
-                 gamma=1.,
-                 scale_fn=None,
-                 scale_mode='cycle',
-                 cycle_momentum=True,
-                 base_momentum=0.8,
-                 max_momentum=0.9,
-                 step_on_iteration=True):
+                 base_lr: float,
+                 max_lr: float,
+                 step_size_up: int = 2000,
+                 step_size_down: Optional[int] = None,
+                 mode: str = 'triangular',
+                 gamma: float = 1.,
+                 scale_fn: Optional[Callable[[float], float]] = None,
+                 scale_mode: str = 'cycle',
+                 cycle_momentum: bool = True,
+                 base_momentum: float = 0.8,
+                 max_momentum: float = 0.9,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = True):
         super().__init__(
             lambda opt: _scheduler.CyclicLR(opt,
                                             base_lr,
@@ -257,7 +296,8 @@ class CyclicLR(LRScheduler):
                                             scale_mode=scale_mode,
                                             cycle_momentum=cycle_momentum,
                                             base_momentum=base_momentum,
-                                            max_momentum=max_momentum),
+                                            max_momentum=max_momentum,
+                                            last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -272,21 +312,24 @@ class CosineAnnealingWarmRestarts(LRScheduler):
         T_0 (int): Number of epochs or iterations for the first restart.
         T_mult (int): T increase factor after a restart.
         eta_min (float, optional): Min learning rate. Defaults to 0.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
 
     """
 
     def __init__(self,
-                 T_0,
-                 T_mult=1,
-                 eta_min=0,
-                 step_on_iteration=False):
+                 T_0: int,
+                 T_mult: int = 1,
+                 eta_min: int = 0,
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         super().__init__(
             lambda opt: _scheduler.CosineAnnealingWarmRestarts(opt,
                                                                T_0,
                                                                T_mult=T_mult,
-                                                               eta_min=eta_min),
+                                                               eta_min=eta_min,
+                                                               last_epoch=last_epoch),
             step_on_iteration=step_on_iteration
         )
 
@@ -298,19 +341,26 @@ class MultiplicativeLR(LRScheduler):
     in the specified function.
 
     Args:
-        lr_lambda (function or list): A function which computes a multiplicative
-            factor given an integer parameter epoch, or a list of such
-            functions, one for each group in optimizer.param_groups.
+        lr_lambda (function or list of functions): A function which computes a
+            multiplicative factor given an integer parameter epoch, or a list
+            of such functions, one for each group in optimizer.param_groups.
+        last_epoch (int): The index of last epoch. Default: -1.
         step_on_iteration (bool): Step on each training iteration rather than each epoch.
             Defaults to False.
+
     """
 
-    def __init__(self, lr_lambda, step_on_iteration=False):
+    def __init__(self,
+                 lr_lambda: Union[Callable[[int], float],
+                                  List[Callable[[int], float]]],
+                 last_epoch: int = -1,
+                 step_on_iteration: bool = False):
         from distutils.version import LooseVersion
         if LooseVersion(torch.__version__) >= LooseVersion("1.4.0"):
             super().__init__(
                 lambda opt: _scheduler.MultiplicativeLR(opt,
-                                                        lr_lambda),
+                                                        lr_lambda,
+                                                        last_epoch=last_epoch),
                 step_on_iteration=step_on_iteration
             )
         else:
@@ -327,41 +377,41 @@ class OneCycleLR(LRScheduler):
     than the initial learning rate.
 
     Args:
-        max_lr (float or list): Upper learning rate boundaries in the cycle
-            for each parameter group.
+        max_lr (float or list of float): Upper learning rate boundaries in the
+            cycle for each parameter group.
         total_steps (int): The total number of steps in the cycle. Note that
-            if a value is not provided here, then it must be inferred by providing
-            a value for epochs and steps_per_epoch.
+            if a value is not provided here, then it must be inferred by
+            providing a value for epochs and steps_per_epoch.
             Defaults to None.
         epochs (int): The number of epochs to train for. This is used along
-            with steps_per_epoch in order to infer the total number of steps in the cycle
-            if a value for total_steps is not provided.
+            with steps_per_epoch in order to infer the total number of steps in
+            the cycle if a value for total_steps is not provided.
             Defaults to None.
-        steps_per_epoch (int): The number of steps per epoch to train for. This is
-            used along with epochs in order to infer the total number of steps in the
-            cycle if a value for total_steps is not provided.
+        steps_per_epoch (int): The number of steps per epoch to train for. This
+            is used along with epochs in order to infer the total number of
+            steps in the cycle if a value for total_steps is not provided.
             Defaults to None.
-        pct_start (float): The percentage of the cycle (in number of steps) spent
-            increasing the learning rate.
+        pct_start (float): The percentage of the cycle (in number of steps)
+            spent increasing the learning rate.
             Defaults to 0.3.
         anneal_strategy (str): {'cos', 'linear'}
-            Specifies the annealing strategy: "cos" for cosine annealing, "linear" for
-            linear annealing.
+            Specifies the annealing strategy: "cos" for cosine annealing,
+            "linear" for linear annealing.
             Defaults to 'cos'.
         cycle_momentum (bool): If ``True``, momentum is cycled inversely
             to learning rate between 'base_momentum' and 'max_momentum'.
             Defaults to True.
-        base_momentum (float or list): Lower momentum boundaries in the cycle
-            for each parameter group. Note that momentum is cycled inversely
-            to learning rate; at the peak of a cycle, momentum is
+        base_momentum (float or list of float): Lower momentum boundaries in
+            the cycle for each parameter group. Note that momentum is cycled
+            inversely to learning rate; at the peak of a cycle, momentum is
             'base_momentum' and learning rate is 'max_lr'.
             Defaults to 0.85.
-        max_momentum (float or list): Upper momentum boundaries in the cycle
-            for each parameter group. Functionally,
+        max_momentum (float or list of float): Upper momentum boundaries in
+            the cycle for each parameter group. Functionally,
             it defines the cycle amplitude (max_momentum - base_momentum).
             Note that momentum is cycled inversely
-            to learning rate; at the start of a cycle, momentum is 'max_momentum'
-            and learning rate is 'base_lr'
+            to learning rate; at the start of a cycle, momentum is
+            'max_momentum' and learning rate is 'base_lr'
             Defaults to 0.95.
         div_factor (float): Determines the initial learning rate via
             initial_lr = max_lr/div_factor
@@ -369,21 +419,23 @@ class OneCycleLR(LRScheduler):
         final_div_factor (float): Determines the minimum learning rate via
             min_lr = initial_lr/final_div_factor
             Defaults to 1e4.
+        last_epoch (int): The index of last epoch. Default: -1.
 
     """
 
     def __init__(self,
-                 max_lr,
-                 total_steps=None,
-                 epochs=None,
-                 steps_per_epoch=None,
-                 pct_start=0.3,
-                 anneal_strategy='cos',
-                 cycle_momentum=True,
-                 base_momentum=0.85,
-                 max_momentum=0.95,
-                 div_factor=25.,
-                 final_div_factor=1e4):
+                 max_lr: Union[float, List[float]],
+                 total_steps: Optional[int] = None,
+                 epochs: Optional[int] = None,
+                 steps_per_epoch: Optional[int] = None,
+                 pct_start: float = 0.3,
+                 anneal_strategy: str = 'cos',
+                 cycle_momentum: bool = True,
+                 base_momentum: Union[float, List[float]] = 0.85,
+                 max_momentum: Union[float, List[float]] = 0.95,
+                 div_factor: float = 25.,
+                 final_div_factor: float = 1e4,
+                 last_epoch: int = -1):
         from distutils.version import LooseVersion
         if LooseVersion(torch.__version__) >= LooseVersion("1.3.0"):
             super().__init__(
@@ -398,7 +450,8 @@ class OneCycleLR(LRScheduler):
                                                   base_momentum=base_momentum,
                                                   max_momentum=max_momentum,
                                                   div_factor=div_factor,
-                                                  final_div_factor=final_div_factor),
+                                                  final_div_factor=final_div_factor,
+                                                  last_epoch=last_epoch),
                 step_on_iteration=True
             )
         else:
