@@ -1,2 +1,2 @@
-from argus.metrics.metric import Metric
+from argus.metrics.metric import Metric, attach_metrics
 from argus.metrics.categorical_accuracy import CategoricalAccuracy
