@@ -1,25 +1,27 @@
-from typing import Callable, Union
-import collections
-import warnings
-import logging
-import copy
 import sys
+import copy
+import logging
+import warnings
+import collections
+from typing import Callable, Union, Any, Type, Dict, Tuple, Iterable
 
 import torch
 from torch import nn
 from torch.optim.optimizer import Optimizer
 from torch.nn.parallel import DataParallel, DistributedDataParallel
 
+import argus
+from argus import types
 from argus.loss import pytorch_losses
 from argus.optimizer import pytorch_optimizers
-from argus.utils import device_to_str, Identity, check_pickleble
+from argus.utils import device_to_str, Identity, check_pickleble, get_device_indices
 
 
 ATTRS_BUILD_ORDER = ('nn_module', 'optimizer', 'loss', 'device', 'prediction_transform')
 TRAIN_ATTRS = {'nn_module', 'optimizer', 'loss', 'device', 'prediction_transform'}
 PREDICT_ATTRS = {'nn_module', 'device', 'prediction_transform'}
 ALL_ATTRS = TRAIN_ATTRS | PREDICT_ATTRS
-MODEL_REGISTRY = {}
+MODEL_REGISTRY: Dict[str, Type['argus.model.Model']] = dict()
 
 DEFAULT_ATTRIBUTE_VALUES = {
     'nn_module': None,
@@ -30,37 +32,37 @@ DEFAULT_ATTRIBUTE_VALUES = {
 }
 
 
-def cast_optimizer(optimizer):
+def cast_optimizer(
+        optimizer: Union[Optimizer, Callable, str]
+) -> Union[Optimizer, Callable]:
     if callable(optimizer):
         return optimizer
     elif isinstance(optimizer, str) and optimizer in pytorch_optimizers:
-        optimizer = getattr(torch.optim, optimizer)
-        return optimizer
+        return getattr(torch.optim, optimizer)
     raise TypeError(f"Incorrect type for optimizer {type(optimizer)}")
 
 
-def cast_nn_module(nn_module):
+def cast_nn_module(nn_module: Union[nn.Module, Callable]) -> Union[nn.Module, Callable]:
     if callable(nn_module):
         return nn_module
     raise TypeError(f"Incorrect type for nn_module {type(nn_module)}")
 
 
-def cast_loss(loss):
+def cast_loss(loss: Union[nn.Module, Callable, str]) -> Union[nn.Module, Callable]:
     if callable(loss):
         return loss
     elif isinstance(loss, str) and loss in pytorch_losses:
-        loss = getattr(nn.modules.loss, loss)
-        return loss
+        return getattr(nn.modules.loss, loss)
     raise TypeError(f"Incorrect type for loss {type(loss)}")
 
 
-def cast_prediction_transform(transform):
+def cast_prediction_transform(transform: Callable) -> Callable:
     if callable(transform):
         return transform
     raise TypeError(f"Incorrect type for prediction_transform: {type(transform)}")
 
 
-def cast_device(device):
+def cast_device(device: types.InputDevices) -> types.Devices:
     if isinstance(device, torch.device):
         return device
     elif isinstance(device, (list, tuple)):
@@ -97,26 +99,33 @@ class ModelMeta(type):
         return new_class
 
 
-def choose_attribute_from_dict(attribute_meta, attribute_params):
-    if isinstance(attribute_meta, collections.Mapping):
+def choose_attribute_from_dict(
+        attribute_meta: types.AttrMeta,
+        attribute_params: types.Param
+) -> Tuple[Any, collections.abc.Mapping]:
+    if isinstance(attribute_meta, collections.abc.Mapping):
         if isinstance(attribute_params, (list, tuple)) and len(attribute_params) == 2:
             name, params = attribute_params
             if name not in attribute_meta:
                 raise ValueError(f"Attribute '{name}' there is not in "
-                                 f"attribute params {attribute_meta}.")
+                                 f"attribute meta '{attribute_meta}'.")
+            if not isinstance(params, collections.abc.Mapping):
+                raise TypeError(f"Attribute params '{params}' should be a "
+                                f"dictionary, not '{type(params)}'.")
         elif isinstance(attribute_params, str):
             name, params = attribute_params, dict()
         else:
-            raise TypeError(f"Incorrect attribute params {attribute_params}")
+            raise TypeError(f"Incorrect attribute params '{attribute_params}' "
+                            f"for attribute meta '{attribute_meta}'. Attribute "
+                            f"params should be str or (str, dict).")
         attribute = attribute_meta[name]
     else:
         attribute = attribute_meta
+        if not isinstance(attribute_params, collections.abc.Mapping):
+            raise TypeError(f"Attribute params '{attribute_params}' should be a "
+                            f"dictionary, not '{type(attribute_params)}'.")
         params = attribute_params
 
-    if not isinstance(params, collections.Mapping):
-        raise TypeError(f"Attribute params should be a dictionary, "
-                        f"not {type(params)}.")
-
     return attribute, params
 
 
@@ -126,8 +135,9 @@ class BuildModel(metaclass=ModelMeta):
     loss: nn.Module
     device: torch.device
     prediction_transform: Callable
+    _meta: Dict[str, types.AttrMeta]
 
-    def __init__(self, params: dict, build_order: list = ATTRS_BUILD_ORDER):
+    def __init__(self, params: dict, build_order: Iterable = ATTRS_BUILD_ORDER):
         params = copy.deepcopy(params)
         check_pickleble(params)
         self.params = params
@@ -143,47 +153,56 @@ class BuildModel(metaclass=ModelMeta):
                 attribute = attr_build_func(attribute_meta, attribute_params)
             setattr(self, attr_name, attribute)
 
-        self.set_device(self.device)
-
-    def build_nn_module(self, nn_module_meta, nn_module_params):
+    def build_nn_module(self,
+                        nn_module_meta: types.AttrMeta,
+                        nn_module_params: types.Param):
         if nn_module_meta is None:
             raise ValueError("nn_module is required attribute for argus.Model")
 
-        nn_module, nn_module_params = choose_attribute_from_dict(nn_module_meta,
-                                                                 nn_module_params)
+        nn_module, params = choose_attribute_from_dict(nn_module_meta,
+                                                       nn_module_params)
         nn_module = cast_nn_module(nn_module)
-        nn_module = nn_module(**nn_module_params)
+        nn_module = nn_module(**params)
         return nn_module
 
-    def build_optimizer(self, optimizer_meta, optim_params):
-        optimizer, optim_params = choose_attribute_from_dict(optimizer_meta,
-                                                             optim_params)
+    def build_optimizer(self,
+                        optimizer_meta: types.AttrMeta,
+                        optim_params: types.Param):
+        optimizer, params = choose_attribute_from_dict(optimizer_meta,
+                                                       optim_params)
         optimizer = cast_optimizer(optimizer)
         grad_params = (param for param in self.nn_module.parameters()
                        if param.requires_grad)
-        optimizer = optimizer(params=grad_params, **optim_params)
+        optimizer = optimizer(params=grad_params, **params)
         return optimizer
 
-    def build_loss(self, loss_meta, loss_params):
-        loss, loss_params = choose_attribute_from_dict(loss_meta,
-                                                       loss_params)
+    def build_loss(self,
+                   loss_meta: types.AttrMeta,
+                   loss_params: types.Param):
+        loss, params = choose_attribute_from_dict(loss_meta,
+                                                  loss_params)
         loss = cast_loss(loss)
-        loss = loss(**loss_params)
+        loss = loss(**params)
         return loss
 
-    def build_prediction_transform(self, transform_meta, transform_params):
-        transform, transform_params = choose_attribute_from_dict(transform_meta,
-                                                                 transform_params)
+    def build_prediction_transform(self,
+                                   transform_meta: types.AttrMeta,
+                                   transform_params: types.Param):
+        transform, params = choose_attribute_from_dict(transform_meta,
+                                                       transform_params)
         transform = cast_prediction_transform(transform)
-        prediction_transform = transform(**transform_params)
+        prediction_transform = transform(**params)
         return prediction_transform
 
-    def build_device(self, device_meta, device_param):
+    def build_device(self,
+                     device_meta: types.InputDevices,
+                     device_param: types.InputDevices) -> torch.device:
         if device_param:
             device = device_param
         else:
             device = device_meta
-        return cast_device(device)
+        self.set_device(cast_device(device))
+        return self.device
 
     def build_logger(self):
         formatter = logging.Formatter('[%(asctime)s][%(levelname)s]: %(message)s')
@@ -199,36 +218,71 @@ class BuildModel(metaclass=ModelMeta):
         return logger
 
     def get_nn_module(self) -> nn.Module:
+        """Get nn_module without :class:`torch.nn.DataParallel` or
+        :class:`torch.nn.parallel.DistributedDataParallel`.
+
+        Returns:
+            :class:`torch.nn.Module`: nn_module without DP and DDP.
+
+        """
         if isinstance(self.nn_module, (DataParallel, DistributedDataParallel)):
             return self.nn_module.module
         else:
             return self.nn_module
 
-    def set_device(self, device: Union[str, torch.device]):
-        device = cast_device(device)
-        str_device = device_to_str(device)
+    def set_device(self, device: types.InputDevices):
+        """Move nn_module and loss to the specified device.
+
+        If a list of devices is passed, :class:`torch.nn.DataParallel` will be
+        used. Batch tensors will be scattered on dim 0. The first device in the
+        list is the location of the output. By default, device "cuda" is the
+        GPU training on :func:`torch.cuda.current_device`.
+
+        Example:
+
+            .. code-block:: python
+
+                model.set_device("cuda")
+                model.set_device(torch.device("cuda"))
+
+                model.set_device("cuda:0")
+                model.set_device(["cuda:2", "cuda:3"])  # Use DataParallel
+
+                model.set_device([torch.device("cuda:2"),
+                                  torch.device("cuda", index=3)])
+
+        Args:
+            device (str, torch.device or list of devices): A device or list of
+                devices.
+
+        """
+        torch_device = cast_device(device)
         nn_module = self.get_nn_module()
 
-        if isinstance(device, (list, tuple)):
-            device_ids = []
-            for dev in device:
-                if dev.type != 'cuda':
-                    raise ValueError("Non cuda device in list of devices")
-                if dev.index is None:
-                    raise ValueError("Cuda device without index in list of devices")
-                device_ids.append(dev.index)
-            if len(device_ids) != len(set(device_ids)):
-                raise ValueError("Cuda device indices must be unique")
+        if isinstance(torch_device, (list, tuple)):
+            device_ids = get_device_indices(torch_device)
             nn_module = DataParallel(nn_module, device_ids=device_ids)
-            device = device[0]
+            output_device = torch_device[0]
+        else:
+            output_device = torch_device
 
-        self.nn_module = nn_module.to(device)
+        self.nn_module = nn_module.to(output_device)
         if self.loss is not None:
-            self.loss = self.loss.to(device)
-        self.params['device'] = str_device
-        self.device = device
+            self.loss = self.loss.to(output_device)
+        self.params['device'] = device_to_str(torch_device)
+        self.device = output_device
 
-    def _check_attributes(self, attrs) -> bool:
+    def get_device(self) -> types.Devices:
+        """Get device or list of devices in case of multi-GPU mode.
+
+        Returns:
+            torch.device or list of torch.device: A device or list of
+            devices.
+
+        """
+        return cast_device(self.params['device'])
+
+    def _check_attributes(self, attrs: Iterable) -> bool:
         for attr_name in attrs:
             attr_value = getattr(self, attr_name, None)
             if attr_value is None:
@@ -253,5 +307,5 @@ class BuildModel(metaclass=ModelMeta):
                 f"Not all required prediction attributes are there: {PREDICT_ATTRS}"
             )
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return str(self.__dict__)
