@@ -1,47 +1,84 @@
-import os
 import numbers
-from pathlib import Path
-from typing import Iterable, Optional, List, Union, Dict
+from typing import Iterable, Optional, Union, Dict, List
 
 import torch
 
-from argus.model.build import BuildModel, MODEL_REGISTRY, cast_device
+import argus
+from argus import types
+from argus.model.build import BuildModel
 from argus.engine import Engine, State
-from argus.callbacks import Callback, on_epoch_complete
-from argus.callbacks.logging import metrics_logging
-from argus.metrics.metric import Metric, METRIC_REGISTRY
+from argus.callbacks import Callback, attach_callbacks
+from argus.callbacks.logging import default_logging
+from argus.metrics.metric import Metric, attach_metrics
 from argus.metrics.loss import Loss
-from argus.utils import deep_to, deep_detach, device_to_str, default, identity
-
-
-def _attach_callbacks(engine, callbacks):
-    if callbacks is not None:
-        for callback in callbacks:
-            if isinstance(callback, Callback):
-                callback.attach(engine)
-            else:
-                raise TypeError(
-                    f"Expected callback type {Callback}, got {type(callback)}"
-                )
-
-
-def _attach_metrics(engine, metrics):
-    for metric in metrics:
-        if isinstance(metric, str):
-            if metric in METRIC_REGISTRY:
-                metric = METRIC_REGISTRY[metric]()
-            else:
-                raise ValueError(f"Metric '{metric}' not found in scope")
-        if isinstance(metric, Metric):
-            metric.attach(engine)
-        else:
-            raise TypeError(
-                f"Expected metric type {Metric} or str, got {type(metric)}"
-            )
+from argus.utils import deep_to, deep_detach
 
 
 class Model(BuildModel):
-    """Model
+    """Argus model is an abstraction of a trainer/predictor that uses:
+
+    Attributes:
+        nn_module (torch.nn.Module): PyTorch neural network as
+            :class:`torch.nn.Module`.
+        optimizer (torch.optim.Optimizer): Optimizer as
+            :class:`torch.optim.Optimizer`.
+        loss (torch.nn.Module): Loss function as :class:`torch.nn.Module`.
+        device (torch.device): device as :class:`torch.torch.device`.
+            The attribute stores a device location of output in the case of
+            multi-GPU mode. To get all devices use
+            :meth:`argus.model.Model.get_device` method.
+        prediction_transform (Callable): postprocessing function of predictions
+            as :class:`Callable` function or object.
+
+    Args:
+        params (dict): A model parameters.
+
+    Examples:
+
+        One can use several ways to initialize :class:`argus.model.Model`:
+
+        1. Set parameters for each part of the model directly:
+
+        .. code-block:: python
+
+            class MnistModel(argus.Model):
+                nn_module = Net  # torch.nn.Module
+                optimizer = torch.optim.SGD
+                loss = torch.nn.CrossEntropyLoss
+
+            params = {
+                'nn_module': {'n_classes': 10, 'p_dropout': 0.1},
+                'optimizer': {'lr': 0.01},
+                'device': 'cpu'
+            }
+
+            model = MnistModel(params)
+
+        2. Set components of the model from multiple options using two
+        elements tuples (component name, the component init arguments):
+
+        .. code-block:: python
+
+            from torchvision.models import resnet18
+
+            class FlexModel(argus.Model):
+                nn_module = {
+                    'net': Net,
+                    'resnet18': resnet18
+                }
+
+            params = {
+                'nn_module': ('resnet18', {
+                    'pretrained': False,
+                    'num_classes': 1
+                }),
+                'optimizer': ('Adam', {'lr': 0.01}),
+                'loss': 'CrossEntropyLoss',
+                'device': 'cuda'
+            }
+
+            model = FlexModel(params)
+
     """
 
     def __init__(self, params: dict):
@@ -51,13 +88,13 @@ class Model(BuildModel):
         """Perform a single train step.
 
         The method is used by :class:`argus.engine.Engine`.
-        The train step includes input and target tensor transition to the model
-        device, forward pass, loss evaluation, backward pass, and the train
-        batch prediction preparation with a *prediction_transform*.
+        The train step includes input and target tensor transferring to the
+        model device, forward pass, loss evaluation, backward pass, and the
+        train batch prediction treating with a prediction_transform.
 
         Args:
-            batch (tuple of 2 torch.Tensors: (input, target)): The input data
-                and target tensors to process.
+            batch (tuple of 2 torch.Tensors: (input, target)): The input and
+                target tensors to process.
             state (:class:`argus.engine.State`): The argus model state.
 
         Returns:
@@ -66,7 +103,7 @@ class Model(BuildModel):
                 {
                     'prediction': The train batch predictions,
                     'target': The train batch target data on the model device,
-                    'loss': Loss function value
+                    'loss': The loss function value
                 }
 
         """
@@ -91,11 +128,11 @@ class Model(BuildModel):
         """Perform a single validation step.
 
         The method is used by :class:`argus.engine.Engine`.
-        The validation step includes input and target tensor transition to the
-        model device, forward pass, loss evaluation, and the train batch
-        prediction preparation with a *prediction_transform*.
+        The validation step includes input and target tensor transferring to
+        the model device, forward pass, loss evaluation, and the val batch
+        prediction treating with a prediction_transform.
 
-        Gradient calculations and the model weights update are omitted, which
+        Gradients calculation and the model weights update are omitted, which
         is the main difference with the :meth:`train_step`
         method.
 
@@ -105,18 +142,19 @@ class Model(BuildModel):
             state (:class:`argus.engine.State`): The argus model state.
 
         Returns:
-            dict: The train step results::
+            dict: Default val step results::
 
                 {
-                    'prediction': The train batch predictions,
-                    'target': The train batch target data on the model device,
-                    'loss': Loss function value
+                    'prediction': The val batch predictions,
+                    'target': The val batch target data on the model device,
+                    'loss': The loss function value
                 }
 
         """
         self.eval()
         with torch.no_grad():
-            input, target = deep_to(batch, device=self.device, non_blocking=True)
+            input, target = deep_to(
+                batch, device=self.device, non_blocking=True)
             prediction = self.nn_module(input)
             loss = self.loss(prediction, target)
             prediction = self.prediction_transform(prediction)
@@ -130,22 +168,22 @@ class Model(BuildModel):
             train_loader: Iterable,
             val_loader: Optional[Iterable] = None,
             num_epochs: int = 1,
-            metrics: Optional[List[Union[Metric, str]]] = None,
+            metrics: List[Union[Metric, str]] = None,
             metrics_on_train: bool = False,
             callbacks: Optional[List[Callback]] = None,
             val_callbacks: Optional[List[Callback]] = None):
         """Train the argus model.
 
-        The method attaches metrics and callbacks to the train and validation,
-        and runs the training process.
+        The method attaches metrics and callbacks to the train and validation
+        process, and performs training itself.
 
         Args:
             train_loader (Iterable): The train data loader.
-            val_loader (Iterable, optional):
-                The validation data loader. Defaults to `None`.
+            val_loader (Iterable, optional): The validation data loader.
+                Defaults to `None`.
             num_epochs (int, optional): Number of training epochs to
                 run. Defaults to 1.
-            metrics (list of :class:`argus.metrics.Metric`, optional):
+            metrics (list of :class:`argus.metrics.Metric` or str, optional):
                 List of metrics to evaluate. By default, the metrics are
                 evaluated on the validation data (if any) only.
                 Defaults to `None`.
@@ -162,40 +200,38 @@ class Model(BuildModel):
         self._check_train_ready()
         metrics = [] if metrics is None else metrics
 
-        train_engine = Engine(self.train_step, model=self,
-                              logger=self.logger, phase='train')
+        train_engine = Engine(self.train_step)
         train_metrics = [Loss()] + metrics if metrics_on_train else [Loss()]
-        _attach_metrics(train_engine, train_metrics)
-        metrics_logging.attach(train_engine, train=True)
+        attach_metrics(train_engine, train_metrics)
+        default_logging.attach(train_engine)
 
         if val_loader is not None:
             self.validate(val_loader, metrics, val_callbacks)
-            val_engine = Engine(self.val_step, model=self,
-                                logger=self.logger, phase='val')
-            _attach_metrics(val_engine, [Loss()] + metrics)
-            _attach_callbacks(val_engine, val_callbacks)
+            val_engine = Engine(self.val_step)
+            attach_metrics(val_engine, [Loss()] + metrics)
+            default_logging.attach(val_engine)
+            attach_callbacks(val_engine, val_callbacks)
 
-            @on_epoch_complete
+            @argus.callbacks.on_epoch_complete
             def validation_epoch(train_state, val_engine, val_loader):
                 epoch = train_state.epoch
                 val_state = val_engine.run(val_loader, epoch, epoch + 1)
                 train_state.metrics.update(val_state.metrics)
 
             validation_epoch.attach(train_engine, val_engine, val_loader)
-            metrics_logging.attach(train_engine, train=False)
 
-        _attach_callbacks(train_engine, callbacks)
+        attach_callbacks(train_engine, callbacks)
         train_engine.run(train_loader, 0, num_epochs)
 
     def validate(self,
-                 val_loader: Optional[Iterable],
-                 metrics: Optional[List[Metric]] = None,
+                 val_loader: Iterable,
+                 metrics: Optional[List[Union[Metric, str]]] = None,
                  callbacks: Optional[List[Callback]] = None) -> Dict[str, float]:
         """Perform a validation.
 
         Args:
             val_loader (Iterable): The validation data loader.
-            metrics (list of :class:`argus.metrics.Metric`, optional):
+            metrics (list of :class:`argus.metrics.Metric` or str, optional):
                 List of metrics to evaluate with the data. Defaults to `None`.
             callbacks (list of :class:`argus.callbacks.Callback`, optional):
                 List of callbacks to be attached to the validation process.
@@ -207,12 +243,12 @@ class Model(BuildModel):
         """
         self._check_train_ready()
         metrics = [] if metrics is None else metrics
-        val_engine = Engine(self.val_step, model=self,
-                            logger=self.logger, phase='val')
-        _attach_metrics(val_engine, [Loss()] + metrics)
-        _attach_callbacks(val_engine, callbacks)
-        metrics_logging.attach(val_engine, train=False, print_epoch=False)
-        return val_engine.run(val_loader).metrics
+        val_engine = Engine(self.val_step)
+        attach_metrics(val_engine, [Loss()] + metrics)
+        default_logging.attach(val_engine)
+        attach_callbacks(val_engine, callbacks)
+        state = val_engine.run(val_loader, -1, 0)
+        return state.metrics
 
     def set_lr(self, lr: Union[float, List[float]]):
         """Set the learning rate for the optimizer.
@@ -231,7 +267,7 @@ class Model(BuildModel):
         Raises:
             ValueError: If *lr* is a list or tuple and its length is not equal
                 to the number of parameter groups.
-            ValueError: If *lr* type is not list, tuple, or number.
+            ValueError: If *lr* type is not a list, tuple, or number.
             AttributeError: If the model is not *train_ready* (i.e. not all
                 attributes are set).
 
@@ -258,7 +294,7 @@ class Model(BuildModel):
         parameter groups.
 
         Returns:
-            (float or a list of floats): The learning rate value or a list of
+            float or a list of float: The learning rate value or a list of
             individual parameter groups learning rate values.
 
         """
@@ -270,7 +306,7 @@ class Model(BuildModel):
             return lrs[0]
         return lrs
 
-    def save(self, file_path: Union[str, Path]):
+    def save(self, file_path: types.Path, optimizer_state: bool = False):
         """Save the argus model into a file.
 
         The argus model is saved as a dict::
@@ -278,13 +314,16 @@ class Model(BuildModel):
             {
                 'model_name': Name of the argus model,
                 'params': Argus model parameters dict,
-                'nn_state_dict': torch nn_module.state_dict()
+                'nn_state_dict': torch nn_module.state_dict(),
+                'optimizer_state_dict': torch optimizer.state_dict()
             }
 
-        The *state_dict* is always transferred to cpu prior to saving.
+        The *state_dict* is always transferred to CPU before saving.
 
         Args:
-            file_path (str): Path to the argus model file.
+            file_path (str or :class:`pathlib.Path`): Path to the argus model
+                file.
+            optimizer_state (bool): Save optimizer state. Defaults to False.
 
         """
         nn_module = self.get_nn_module()
@@ -293,6 +332,10 @@ class Model(BuildModel):
             'params': self.params,
             'nn_state_dict': deep_to(nn_module.state_dict(), 'cpu')
         }
+        if optimizer_state and self.optimizer is not None:
+            state['optimizer_state_dict'] = deep_to(
+                self.optimizer.state_dict(), 'cpu'
+            )
         torch.save(state, file_path)
         self.logger.info(f"Model saved to '{file_path}'")
 
@@ -310,7 +353,7 @@ class Model(BuildModel):
 
         Returns:
             torch.Tensor or other type: Predictions as the result of the
-                prediction_transform application.
+            prediction_transform application.
 
         """
         self._check_predict_ready()
@@ -330,96 +373,3 @@ class Model(BuildModel):
         """Set the nn_module into eval mode."""
         if self.nn_module.training:
             self.nn_module.eval()
-
-
-def load_model(file_path: Union[str, Path],
-               nn_module=default,
-               optimizer=default,
-               loss=default,
-               prediction_transform=default,
-               device=default,
-               change_params_func=identity,
-               change_state_dict_func=identity,
-               model_name=default,
-               **kwargs):
-    """Load an argus model from a file.
-
-    The function allows loading an argus model, saved with
-    :meth:`argus.model.Model.save`. The model is always loaded in *eval* mode.
-
-    Args:
-        file_path (str): Path to the file to load.
-        device (str or :class:`torch.device`, optional): Device for the model.
-            Defaults to None.
-        nn_module (dict, tuple or str, optional): Params of the nn_module to
-            replace params in the state.
-        optimizer (dict, tuple or str, optional): Params of the optimizer to
-            replace params in the state. Set to `None` if don't want to create
-            optimizer in the loaded model.
-        loss (dict, tuple or str, optional): Params of the loss to replace params
-            in the state. Set to `None` if don't want to create loss in the
-            loaded model.
-        prediction_transform (dict, tuple or str, optional): Params of the
-            prediction_transform to replace params in the state. Set to `None`
-            if don't want to create prediction_transform in the loaded model.
-        change_params_func (function, optional): Function for modification of
-            state params. Takes as input params from the loaded state, outputs
-            params to model creation.
-        change_state_dict_func (function, optional): Function for modification of
-            nn_module state dict. Takes as input state dict from the loaded
-            state, outputs state dict to model creation.
-        model_name (str): Class name of :class:`argus.model.Model`.
-            By default uses name from loaded state.
-
-    Raises:
-        ImportError: If the model is not available in the scope. Often it means
-            that it is not imported or defined.
-        FileNotFoundError: If the file is not found by the *file_path*.
-
-    Returns:
-        :class:`argus.model.Model`: Loaded argus model.
-
-    """
-
-    if os.path.isfile(file_path):
-        state = torch.load(file_path)
-
-        if model_name is default:
-            model_name = state['model_name']
-        else:
-            model_name = model_name
-
-        if model_name in MODEL_REGISTRY:
-            params = state['params']
-            if device is not default:
-                device = cast_device(device)
-                device = device_to_str(device)
-                params['device'] = device
-
-            if nn_module is not default:
-                if nn_module is None:
-                    raise ValueError("nn_module is required attribute for argus.Model")
-                params['nn_module'] = nn_module
-            if optimizer is not default:
-                params['optimizer'] = optimizer
-            if loss is not default:
-                params['loss'] = loss
-            if prediction_transform is not default:
-                params['prediction_transform'] = prediction_transform
-
-            for attribute, attribute_params in kwargs.items():
-                params[attribute] = attribute_params
-
-            model_class = MODEL_REGISTRY[model_name]
-            params = change_params_func(params)
-            model = model_class(params)
-            nn_state_dict = deep_to(state['nn_state_dict'], model.device)
-            nn_state_dict = change_state_dict_func(nn_state_dict)
-
-            model.get_nn_module().load_state_dict(nn_state_dict)
-            model.eval()
-            return model
-        else:
-            raise ImportError(f"Model '{model_name}' not found in scope")
-    else:
-        raise FileNotFoundError(f"No state found at {file_path}")
